#include "GLWidget.h"
#include <algorithm>
#include <cstdlib>
#include <cstdio>
#include <qtimer.h>

#include <qmessagebox.h>

//Types values
const unsigned int H_FLUID = 0;
const unsigned int H_NOSLIP = 1;
const unsigned int H_VELOCITY = 2;
const unsigned int H_OPENBOUNDARY_N = 3;
const unsigned int H_OPENBOUNDARY_S = 4;
const unsigned int H_OPENBOUNDARY_E = 5;
const unsigned int H_OPENBOUNDARY_W = 6;
const unsigned int H_OPENBOUNDARY_VEL = 7;

//Rang of values to calculate color simulation
const float MAX_VAL = 1.33f;
const float MIN_VAL = 0.93f;

//Values to draw the cubes
float cube_edge;
float cubes_offset;

//Buffer data to draw the scene
static GLfloat *vertices;
static GLfloat *colors;
static GLuint *indices;

int indices_length;

//To avoid initial errors
bool first_exec = true;

GLWidget::GLWidget(QWidget *parent)
	: QGLWidget(parent)
{
	qtPurple = QColor::fromCmykF(0.39, 0.39, 0.0, 0.0);
	filename = QString(QLatin1String(""));
	simulate = false;
	iter = 0;
	vertices = new GLfloat[0];
	colors = new GLfloat[0];
	indices = new GLuint[0];
	indices_length = 0;

	wind_dir = new QString("N");

	QTimer *timer = new QTimer(this);
	connect(timer, SIGNAL(timeout()), this, SLOT(update()));
	timer->start(1);
}

GLWidget::~GLWidget()
{
	//Delete data
	delete [] vertices;
	delete [] colors;
	delete [] indices;
	
	delete image;
	/* Error al eliminar el simulador...
	delete sim;
	*/
}

QSize GLWidget::minimumSizeHint() const
{
	return QSize(50, 50);
}

QSize GLWidget::sizeHint() const
{
	return QSize(2000, 2000);
}

void GLWidget::setFilename(QString fn)
{
	filename = fn;
}

QString GLWidget::getFilename()
{
	return filename;
}

void GLWidget::loadScene()
{
	//Delete previous image and simulation
	if(first_exec) { //In first time no data would be allocated in the pointer, so we avoid to delete null pointers
		first_exec = false;		
	} else { 
		delete image;
		//delete sim;
	}

	iter = 0;

	//Read new image
	//To read the image, we need to cast the QString to char*
	QByteArray ba = filename.toLatin1(); //First we convert to QByteArray
	char *image_path = ba.data();		//Then we extract the pointer to the data
	image = ir.readImage(image_path);

	scene_width = image->getWidth();
	scene_height = image->getHeight();

	//Create new simulator
	sim = new Simulator(scene_width, scene_height);
	sim_types = sim->initialise(image);
	//Set simulation default wind direction
	if(wind_dir == QString("N")) sim->setWindDirection(0.f,1.f);
	else if(wind_dir == QString("NE")) sim->setWindDirection(1.f,1.f);
	else if(wind_dir == QString("E")) sim->setWindDirection(1.f,0.f);
	else if(wind_dir == QString("SE")) sim->setWindDirection(1.f,-1.f);
	else if(wind_dir == QString("S")) sim->setWindDirection(0.f,-1.f);
	else if(wind_dir == QString("SW")) sim->setWindDirection(-1.f,-1.f);
	else if(wind_dir == QString("W")) sim->setWindDirection(-1.f,0.f);
	else if(wind_dir == QString("NW")) sim->setWindDirection(-1.f,1.f);
	else printf("Error, bad wind direction\n");

	//Init sim data
	sim_velocities = new float[scene_width * scene_height *2];
	sim_densities = new float[scene_width * scene_height];

	//Prepare display data
	int width, height;
	width = this->width();
	height = this->height();
	//QMessageBox::information(this, tr("Console"), "WidgetGL width: %i, height: %i\n");
	float width2, height2;
	width2 = 0.99f;
	height2 = 0.99f;

	//float edgeX = ((float)width2 / (1.1f*scene_width)) / ((float)width2/2.f);
	float edgeX = width2 / (float)scene_width;
	//float edgeY = ((float)height2 / (1.1f*scene_height)) / ((float)height2/2.f);
	float edgeY = height2 / (float)scene_height;
	cube_edge = std::min(edgeX, edgeY);
	cubes_offset = 0.1f * cube_edge;
	cubes_offset = 0.f; //!!!!

	//Initialise data info to draw
	initVertices(scene_width, scene_height);
	initColors(scene_width, scene_height, sim_types); //By default we'll init colors to white, later in each iteration it will be update
	initIndices(scene_width, scene_height);

	//Load vertex and color info
	glColorPointer(3, GL_FLOAT, 0, colors);
	glVertexPointer(3, GL_FLOAT, 0, vertices);

	//glPointSize(10.f);
	//glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void GLWidget::continueSimulation()
{
	simulate = true;
}

void GLWidget::stopSimulation()
{
	simulate = false;
}

void GLWidget::initializeGL()
{
	QColor *c = new QColor(0.f,0.f,0.f,0.f);
	//qglClearColor(qtPurple.dark());
	qglClearColor(*c);

    glEnable(GL_DEPTH_TEST);
    glEnable(GL_CULL_FACE);
    glShadeModel(GL_SMOOTH);
    /*glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glEnable(GL_MULTISAMPLE);
    static GLfloat lightPosition[4] = { 0.5, 5.0, 7.0, 1.0 };
    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
	*/

	//Enable array structure to draw
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_VERTEX_ARRAY);
}

void GLWidget::paintGL()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    glTranslatef(0.0, 0.0, -10.0);

	if(simulate) {
		if(sim->nextIteration(sim_velocities, sim_densities) != 0) { //De moment no utilitzarem les velocitats, ms endavant si hi ha temps
			printf("Error in iteration %d of simumlation", iter);
			//return 0;
		}
		//QMessageBox::information(this, tr("Console"), "Next iteration fo simulation executed\n");
		//writeDensities(iter, scene_width, scene_height, sim->getDensities());
		//Recalculate colors of simulation and draw all scene again
		updateColors(sim_types, scene_width, scene_height, sim_densities);
		//QMessageBox::information(this, tr("Console"), "Colors updated\n");
		iter++;
	}
	
	/*glColor3f(0,1,1);
    glBegin(GL_TRIANGLES);
		glVertex2f(0.5f,0.1f);
		glVertex2f(0.1f,0.5f);
		glVertex2f(0.f,0.f);
    glEnd();*/
	//if(simulate) glDrawElements(GL_TRIANGLES, 36*scene_width*scene_height, GL_UNSIGNED_INT, indices);
	glDrawElements(GL_TRIANGLES, indices_length, GL_UNSIGNED_INT, indices);
	//swapBuffers();
}

void GLWidget::resizeGL(int width, int height)
{
	int side = qMin(width, height);
    glViewport((width - side) / 2, (height - side) / 2, side, side);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
#ifdef QT_OPENGL_ES_1
    glOrthof(-0.5, +0.5, -0.5, +0.5, 4.0, 15.0);
#else
    glOrtho(-0.5, +0.5, -0.5, +0.5, 4.0, 15.0);
#endif
    glMatrixMode(GL_MODELVIEW);
}

void GLWidget::setWindDirection(QString dir)
{
	//Stop the simulation to update values
	//simulate = false;
	if(dir != *wind_dir) {
		if(dir == QString("N")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(0.f,1.f);
		} else if(dir == QString("NE")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(1.f,1.f);
		} else if(dir == QString("E")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(1.f,0.f);
		} else if(dir == QString("SE")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(1.f,-1.f);
		} else if(dir == QString("S")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(0.f,-1.f);
		} else if(dir == QString("SW")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(-1.f,-1.f);
		} else if(dir == QString("W")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(-1.f,0.f);
		} else if(dir == QString("NW")) {
			wind_dir = new QString(dir);
			sim->setWindDirection(-1.f,1.f);
		} else {
			QMessageBox::information(this, tr("Error"), "Bad wind direction");
		}
	}
	
	//Continue the simulation
	//simulate = true;
}

//========================================================================
//	Init vertex data
//
//	Cube struct
//		7-----------6
//		|\			|\
//		| 3-----------2
//		| |			| |
//		4-|---------5 |
//		 \|			 \|
//		  0-----------1
//
//========================================================================
void GLWidget::initVertices(int w, int h) {
	vertices = new GLfloat[8*3*w*h]; 
	
	//Calculate the init position, it'll be top left
	//float initX = -1.f * ( ((float) w/2.f) * cube_edge + (((float) w/2.f) - 0.5f) * cubes_offset);
	float initX = -1.f * ( ((float)w / 2.f) * cube_edge);
	//float initY = ( ((float) h/2.f) * cube_edge + (((float) h/2.f) - 0.5f) * cubes_offset ) - cube_edge;
	float initY = ( ((float)h / 2.f) * cube_edge);
	
	int init_index;
	float posX, posY;
	for(int i=0; i<h; i++) {
		for(int j=0; j<w; j++) {
			init_index = (i*w + j) * 8*3; //Each iteration will write 24 values (8 vertices * 3 components x,y,z), so we need to know were start to write
			//Calculate coord for vertex 0 of current cube
			posX = initX + j*(cube_edge + cubes_offset);
			posY = initY - i*(cube_edge + cubes_offset);
			
			vertices[init_index + 0] = posX;				vertices[init_index + 1]  = posY;				vertices[init_index + 2] = 0.f;	 //Vertex 0
			vertices[init_index + 3] = posX + cube_edge;	vertices[init_index + 4]  = posY;				vertices[init_index + 5] = 0.f;  //Vertex 1
			vertices[init_index + 6] = posX + cube_edge;	vertices[init_index + 7]  = posY + cube_edge;	vertices[init_index + 8] = 0.f;  //Vertex 2
			vertices[init_index + 9] = posX;				vertices[init_index + 10] = posY + cube_edge;	vertices[init_index + 11] = 0.f; //Vertex 3

			vertices[init_index + 12] = posX;				vertices[init_index + 13] = posY;				vertices[init_index + 14] = -cube_edge; //Vertex 4
			vertices[init_index + 15] = posX + cube_edge;	vertices[init_index + 16] = posY;				vertices[init_index + 17] = -cube_edge; //Vertex 5
			vertices[init_index + 18] = posX + cube_edge;	vertices[init_index + 19] = posY + cube_edge;	vertices[init_index + 20] = -cube_edge; //Vertex 6
			vertices[init_index + 21] = posX;				vertices[init_index + 22] = posY + cube_edge;	vertices[init_index + 23] = -cube_edge; //Vertex 7
		}
	}
}

//========================================================================
//	Init color data
//========================================================================
void GLWidget::initColors(int w, int h, unsigned int *types) {
	float red, green, blue;
	
	colors = new GLfloat[8*3*w*h];
	
	int init_index;
	for(int i=0; i<h; i++) {
		for(int j=0; j<w; j++) {
			init_index = (i*w + j) * 8*3; //Each iteration will write 24 values (8 vertices.color * 3 components x,y,z), so we need to know were start to write

			switch(types[i*w + j]) {
				case H_FLUID:
					red = 0.f; green = 0.f;	blue = 1.f;
					break;
				case H_NOSLIP:
					red = 1.f; green = 0.99216f; blue = 0.81569f;
					break;
				case H_VELOCITY:
					red = 1.f; green = 0.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_N:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_S:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_E:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_W:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_VEL:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				default:
					break;
				}
			//						 Red							Green						   Blue
			colors[init_index + 0] = red; colors[init_index + 1]  = green; colors[init_index + 2]  = blue; //Color for vertex 0
			colors[init_index + 3] = red; colors[init_index + 4]  = green; colors[init_index + 5]  = blue; //Color for vertex 1
			colors[init_index + 6] = red; colors[init_index + 7]  = green; colors[init_index + 8]  = blue; //Color for vertex 2
			colors[init_index + 9] = red; colors[init_index + 10] = green; colors[init_index + 11] = blue; //Color for vertex 3

			colors[init_index + 12] = red; colors[init_index + 13] = green; colors[init_index + 14] = blue; //Color for vertex 4
			colors[init_index + 15] = red; colors[init_index + 16] = green; colors[init_index + 17] = blue; //Color for vertex 5
			colors[init_index + 18] = red; colors[init_index + 19] = green; colors[init_index + 20] = blue; //Color for vertex 6
			colors[init_index + 21] = red; colors[init_index + 22] = green; colors[init_index + 23] = blue; //Color for vertex 7
		}
	}
}

//========================================================================
//	Init index data
//========================================================================
void GLWidget::initIndices(int w, int h) {
	indices = new GLuint[6*2*3*w*h]; //We got 6 faces, each face 2 triangle and each triangle 3 vertices
	indices_length = 6*2*3*w*h;

	int init_index, init_vertex;
	for(int i=0; i<h; i++) {
		for(int j=0; j<w; j++) {
			init_index = (i*w + j) * 6*2*3;//Each iteration will write 32 values (6 faces * 2 trianglePerFace * 3 verticesPerTriangle), so we need to know were start to write
			init_vertex = (i*w + j) * 8;//Each cube works with 8 vertices, vertices[(0,1,2,3,4,5,6,7)cube0,(,8,9,10,11,12,13,14,15)cube1,...]
			
			indices[init_index + 0] = init_vertex + 0; indices[init_index + 1] = init_vertex + 1; indices[init_index + 2] = init_vertex + 3;
			indices[init_index + 3] = init_vertex + 0; indices[init_index + 4] = init_vertex + 3; indices[init_index + 5] = init_vertex + 4;
			indices[init_index + 6] = init_vertex + 0; indices[init_index + 7] = init_vertex + 4; indices[init_index + 8] = init_vertex + 1;

			indices[init_index + 9]  = init_vertex + 2; indices[init_index + 10] = init_vertex + 3; indices[init_index + 11] = init_vertex + 1;
			indices[init_index + 12] = init_vertex + 2; indices[init_index + 13] = init_vertex + 1; indices[init_index + 14] = init_vertex + 6;
			indices[init_index + 15] = init_vertex + 2; indices[init_index + 16] = init_vertex + 6; indices[init_index + 17] = init_vertex + 3;

			indices[init_index + 18] = init_vertex + 5; indices[init_index + 19] = init_vertex + 1; indices[init_index + 20] = init_vertex + 4;
			indices[init_index + 21] = init_vertex + 5; indices[init_index + 22] = init_vertex + 6; indices[init_index + 23] = init_vertex + 1;
			indices[init_index + 24] = init_vertex + 5; indices[init_index + 25] = init_vertex + 4; indices[init_index + 26] = init_vertex + 6;

			indices[init_index + 27] = init_vertex + 7; indices[init_index + 28] = init_vertex + 4; indices[init_index + 29] = init_vertex + 3;
			indices[init_index + 30] = init_vertex + 7; indices[init_index + 31] = init_vertex + 3; indices[init_index + 32] = init_vertex + 6;
			indices[init_index + 33] = init_vertex + 7; indices[init_index + 34] = init_vertex + 6; indices[init_index + 35] = init_vertex + 4;
		}
	}
}

//========================================================================
//	Update color data
//========================================================================
void GLWidget::updateColors(unsigned int *types, int w, int h, float *densities){
	int init_index;
	float aux;
	float red, green, blue;
	for(int i=0; i<h; i++) {
		for(int j=0; j<w; j++) {
			//Calculate RGB values depends on type
			switch(types[i*w + j]) {
				case H_FLUID:
					red = 0.f;
					aux = densities[i*w + j] - MIN_VAL;
					green = (aux/(MAX_VAL-MIN_VAL));
					blue = 1.f;
					break;
				case H_NOSLIP:
					red = 1.f; green = 0.99216f; blue = 0.81569f;
					break;
				case H_VELOCITY:
					red = 1.f; green = 0.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_N:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_S:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_E:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_W:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				case H_OPENBOUNDARY_VEL:
					red = 0.f; green = 1.f; blue = 0.f;
					break;
				default:
					break;
				}
			//As on initColors, we'll update colors buffer depends on type cell
			init_index = (i*w + j) * 8*3;

			//						 Red							Green						   Blue
			colors[init_index + 0] = red; colors[init_index + 1]  = green; colors[init_index + 2]  = blue;
			colors[init_index + 3] = red; colors[init_index + 4]  = green; colors[init_index + 5]  = blue;
			colors[init_index + 6] = red; colors[init_index + 7]  = green; colors[init_index + 8]  = blue;
			colors[init_index + 9] = red; colors[init_index + 10] = green; colors[init_index + 11] = blue;

			colors[init_index + 12] = red; colors[init_index + 13] = green; colors[init_index + 14] = blue;
			colors[init_index + 15] = red; colors[init_index + 16] = green; colors[init_index + 17] = blue;
			colors[init_index + 18] = red; colors[init_index + 19] = green; colors[init_index + 20] = blue;
			colors[init_index + 21] = red; colors[init_index + 22] = green; colors[init_index + 23] = blue;
		}
	}
}

void GLWidget::writeDensities(int iter, int width, int height, float *densities){
	FILE *output;
	char filename[256];
	
	sprintf_s(filename, "outputD_%04d.txt", iter);

	if( (output = fopen(filename, "w")) == NULL ) {
		printf("Unable to open file %s. Ignoring output.", filename);
		return;
	}

	fprintf(output, "P3\n%d %d\n%d\n", width, height, 255);

	for(int i=0; i<height; i++) {
		for(int j=0; j<width; j++) {
			fprintf(output, "%f ", densities[i*width + j]);
		}
		fprintf(output, "*\n");
	}
  
	fclose(output);
}