//Simulatord.cuh implementation file

#include "Simulator.cuh"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

//Helper FUNCTIONS
#define CUDA_SAFE_CALL( call) {										 \
	cudaError err = call;											 \
	if( cudaSuccess != err) {										 \
		fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",\
		__FILE__, __LINE__, cudaGetErrorString( err) );              \
	exit(EXIT_FAILURE);												 \
  } }

// addition
inline __host__ __device__ float2 operator+(float2 a, float2 b) {return make_float2(a.x + b.x, a.y + b.y);}
inline __host__ __device__ float2 operator+(float2 a, float b) {return make_float2(a.x + b, a.y + b);}
inline __host__ __device__ void operator+=(float2 &a, float2 b) {a.x += b.x; a.y += b.y;}
// subtract
inline __host__ __device__ float2 operator-(float2 a, float2 b) {return make_float2(a.x - b.x, a.y - b.y);}
inline __host__ __device__ float2 operator-(float2 a, float b) {return make_float2(a.x - b, a.y - b);}
inline __host__ __device__ void operator-=(float2 &a, float2 b) {a.x -= b.x; a.y -= b.y;}
// multiply
inline __host__ __device__ float2 operator*(float2 a, float2 b) {return make_float2(a.x * b.x, a.y * b.y);}
inline __host__ __device__ float2 operator*(float2 a, float s) {return make_float2(a.x * s, a.y * s);}
inline __host__ __device__ float2 operator*(float s, float2 a) {return make_float2(a.x * s, a.y * s);}
inline __host__ __device__ void operator*=(float2 &a, float s) {a.x *= s; a.y *= s;}
//dot product
inline __host__ __device__ float dot(float2 a, float2 b) {return a.x * b.x + a.y * b.y;}

#define BLOCKSIZE 64

const unsigned int H_FLUID = 0;
const unsigned int H_NOSLIP = 1;
const unsigned int H_VELOCITY = 2;
const unsigned int H_OPENBOUNDARY_N = 3;
const unsigned int H_OPENBOUNDARY_S = 4;
const unsigned int H_OPENBOUNDARY_E = 5;
const unsigned int H_OPENBOUNDARY_W = 6;
const unsigned int H_OPENBOUNDARY_VEL = 7;

__constant__ float weight[9];
__constant__ float2 vectors[9];
__constant__ float inputUx;
__constant__ int inv[9];
__constant__ int next[9];
__constant__ int ref_addr[9];
__constant__ int SIM_WIDTH;
__constant__ int SIM_HEIGHT;

__constant__ unsigned int D_FLUID;
__constant__ unsigned int D_NOSLIP;
__constant__ unsigned int D_VELOCITY;
__constant__ unsigned int D_OPENBOUNDARY_N; //Top
__constant__ unsigned int D_OPENBOUNDARY_S; //Bottom
__constant__ unsigned int D_OPENBOUNDARY_E; //Right side
__constant__ unsigned int D_OPENBOUNDARY_W; //Left side
__constant__ unsigned int D_OPENBOUNDARY_VEL; //BoundaryCondition behind velocity cells

//Lattice Boltzmann Model for the Incompressible Navier-Stokes Equation - Xiaoyi He and Li-Shi Luo
__device__ float equilibrium(float rho, float2 uu, float weight, float2 vec) {
	float uvec = dot(uu, vec);
	return weight * (rho - 1.5f*dot(uu,uu) + 3.f*uvec + 4.5f*uvec*uvec);
}

__global__ void InitLBM(unsigned int *celltype, float *f_0, float2 *uu, float *rho) {
	int pos = threadIdx.x + blockIdx.x * BLOCKSIZE;

	if(pos < SIM_WIDTH * SIM_HEIGHT) {
		int type = celltype[pos];

		if(D_OPENBOUNDARY_N == type || D_OPENBOUNDARY_S == type || D_OPENBOUNDARY_E == type || D_OPENBOUNDARY_W == type || D_OPENBOUNDARY_VEL == type) {
			uu[pos] = make_float2(0.f,0.f);
			rho[pos] = 0.f;
			for(int i=0; i<9; i++) {
				f_0[ref_addr[i] + pos] = equilibrium(1.f, make_float2(0.f,0.f), weight[i], vectors[i]);
				uu[pos] += f_0[ref_addr[i] + pos] * vectors[i];
				rho[pos] += f_0[ref_addr[i] + pos];	
			}
			return;
		 }

		if(D_NOSLIP == type) {
			for(int i=0; i<9; i++) f_0[ref_addr[i] + pos] = 0.f;
			uu[pos] = make_float2(0.f, 0.f);
			rho[pos] = 0.f;
			return; 
			//boundary cells won't be accessed, 
			//but we init the vel to avoid random values on output images
		}
	
		float2 u = make_float2(0.f, 0.f);
		for(int i=0; i<9; i++) {
			f_0[ref_addr[i] + pos] = equilibrium(1.f, u, weight[i], vectors[i]); //Init to rest
		}
	}
}

/*
	ATENCIO: pot haber solapament en els "vertexs"

		WNN...NNE
		W##...##E
		.........
		W##...##E
		WSS...SSE

		el (0,0) W voldra llegir de (1,0) N, pero aquest encara no estar inicialitzat!!!!!
*/
__global__ void InitOpenLBM(unsigned int *celltype, float *f_0, float2 *uu, float *rho) {
	int pos = threadIdx.x + blockIdx.x * BLOCKSIZE;
	
	if(pos < SIM_WIDTH * SIM_HEIGHT) {
		unsigned int ct = celltype[pos];

		if(ct == D_FLUID || ct == D_NOSLIP || ct == D_VELOCITY) return;
		if(ct == D_OPENBOUNDARY_N) {
			if(pos == 0) {
				//Top Left corner
				rho[pos] = rho[(pos + 1) + SIM_WIDTH];
				uu[pos] = uu[(pos + 1) + SIM_WIDTH];
			} else if(pos == (SIM_WIDTH * SIM_HEIGHT - SIM_WIDTH)) {
				//Bottom Left corner
				rho[pos] = rho[(pos + 1) - SIM_WIDTH];
				uu[pos] = uu[(pos + 1) - SIM_WIDTH];
			} else {
				rho[pos] = rho[pos + SIM_WIDTH];
				uu[pos] = uu[pos + SIM_WIDTH];
			}
		} else if(ct == D_OPENBOUNDARY_S) {
			if(pos == 0) {
				//Top Left corner
				rho[pos] = rho[(pos + 1) + SIM_WIDTH];
				uu[pos] = uu[(pos + 1) + SIM_WIDTH];
			} else if(pos == (SIM_WIDTH * SIM_HEIGHT - SIM_WIDTH)) {
				//Bottom Left corner
				rho[pos] = rho[(pos + 1) - SIM_WIDTH];
				uu[pos] = uu[(pos + 1) - SIM_WIDTH];
			} else {
				rho[pos] = rho[pos - SIM_WIDTH];
				uu[pos] = uu[pos - SIM_WIDTH];
			}
		} else if(ct == D_OPENBOUNDARY_E) {
			if(pos == (SIM_WIDTH - 1)) {
				//Top Right corner
				rho[pos] = rho[(pos - 1) + SIM_WIDTH];
				uu[pos] = uu[pos - 1];
			} else if(pos == (SIM_WIDTH * SIM_HEIGHT - 1)) {
				//Bottom Right corner
				rho[pos] = rho[(pos - 1) - SIM_WIDTH];
				uu[pos] = uu[(pos - 1) - SIM_WIDTH];
			} else {
				rho[pos] = rho[pos - 1];
				uu[pos] = uu[pos - 1];
			}
		} else { 
			//D_OPENBOUNDARY_W
			if(pos == 0) {
				//Top Left corner
				rho[pos] = rho[(pos + 1) + SIM_WIDTH];
				uu[pos] = uu[(pos + 1) + SIM_WIDTH];
			} else if(pos == (SIM_WIDTH * SIM_HEIGHT - SIM_WIDTH)) {
				//Bottom Left corner
				rho[pos] = rho[(pos + 1) - SIM_WIDTH];
				uu[pos] = uu[(pos + 1) - SIM_WIDTH];
			} else {
				rho[pos] = rho[pos + 1];
				uu[pos] = uu[pos + 1];
			}
		}

		for(int i=0; i<9; i++) {
			f_0[pos + ref_addr[i]] = equilibrium(rho[pos], uu[pos], weight[i], vectors[i]);
		}
	}
}

__global__ void LBMstep(unsigned int *celltype, float *rho, float2 *u, float *f_0, float *f_1, float omega, float *addWind) {
	int pos = threadIdx.x + blockIdx.x * BLOCKSIZE;

	if(pos < SIM_WIDTH * SIM_HEIGHT) {
		unsigned int ct = celltype[pos];
		if(D_NOSLIP == ct || D_OPENBOUNDARY_N == ct || D_OPENBOUNDARY_S == ct || D_OPENBOUNDARY_E == ct || D_OPENBOUNDARY_W == ct || D_OPENBOUNDARY_VEL == ct) return; //Early exit for boundary cells

		//Stream
		float in_f[9];
	
		in_f[0] = f_0[pos];
		for(int i=1; i<9; i++) {
			//if next cell is boundary, flip actual df
			if(D_NOSLIP == celltype[pos+next[i]]) in_f[i] = f_0[ref_addr[inv[i]] + pos];
			else if(D_OPENBOUNDARY_VEL == celltype[pos+next[i]]) in_f[i] = equilibrium(1.f, make_float2(0.f, 0.f), weight[i], vectors[i]);
			//else stream normal
			else in_f[i] = f_0[ref_addr[i] + pos+next[i]];
		}
	
		__syncthreads();

		//Collide
		if(D_VELOCITY == ct) {
			for(int i=1; i<9; i++) {
				in_f[i] += addWind[i];
			}
		}
		float n_rho = in_f[0];
		float2 n_u = make_float2(0.f, 0.f);

		for(int i=1; i<9; i++) {
			n_rho += in_f[i];
			n_u += in_f[i] * vectors[i];
		}
		//if(D_VELOCITY == ct) n_u.y = inputUx; //Only required for lid-driven cavity example

		for(int i=0; i<9; i++) 
			in_f[i] -= omega * (in_f[i] - equilibrium(n_rho, n_u, weight[i], vectors[i]));
	
		__syncthreads();

		for(int i=0; i<9; i++) 
			f_1[ref_addr[i] + pos] = in_f[i];
  
		rho[pos] = n_rho;
		u[pos] = n_u;
	}
}

//Parameters
const float OMEGA = 1.8f;
const float WIND_PARTICLES = 0.045f;

//LBM D2Q9 constants
const float LBM_W[9] = {(4.f/9.f), 
  (1.f/9.f), (1.f/9.f), (1.f/9.f), (1.f/9.f),
  (1.f/36.f), (1.f/36.f), (1.f/36.f), (1.f/36.f)};
const int LBM_VX[9] = {0, 1, 0, -1,  0, 1, -1, -1,  1};
const int LBM_VY[9] = {0, 0, 1,  0, -1, 1,  1, -1, -1};
const int LBM_INV[9] = {0, 3, 4, 1, 2, 7, 8, 5, 6};

const float2 left_dir = make_float2(1.0,0.0);
const float2 right_dir = make_float2(-1.0,0.0);
const float2 top_dir = make_float2(0.0,-1.0);
const float2 bottom_dir = make_float2(0.0,1.0);

const float2 direction[9] = {make_float2(0.f,0.f), make_float2(1.f,0.f), make_float2(0.f,1.f), make_float2(-1.f,0.f), make_float2(0.f,-1.f),
							make_float2((1.f/sqrt(2.f)),(1.f/sqrt(2.f))), make_float2((-1.f/sqrt(2.f)),(1.f/sqrt(2.f))), 
							make_float2((-1.f/sqrt(2.f)),(-1.f/sqrt(2.f))), make_float2((1.f/sqrt(2.f)),(-1.f/sqrt(2.f)))};

Simulator::Simulator() {
	//NOTHING	
}

Simulator::Simulator(int width, int height) {
	
	//Init attributes available from host
	a_sim_width = width;
	a_sim_height = height;

	//a_inputvel = 0.08f;

	int matdim = a_sim_width * a_sim_height;

	a_f0 = new float[matdim*9];
	a_u = new float2[matdim];
	a_rho = new float[matdim];

	a_ctype = new unsigned int[matdim];

	int nxt_x = 1, nxt_y = a_sim_width;
	for(int i=0; i<9; i++){
		a_lbmvecs[i] = make_float2((float) LBM_VX[i], (float) LBM_VY[i]);
		a_lbmnext[i] = LBM_VX[i]*nxt_x + LBM_VY[i]*nxt_y;
		a_lbmref_addr[i] = i*matdim;
	}

	//Allocating and Moving data to GPU
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_celltype, sizeof(unsigned int) * matdim));
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_rho, sizeof(float) * matdim));
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_f0, sizeof(float) * matdim * 9));
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_f1, sizeof(float) * matdim * 9));
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_u, sizeof(float2) * matdim));
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_addWind, sizeof(float) * matdim * 9));

	//CUDA_SAFE_CALL(cudaMemcpyToSymbol(inputUx, &a_inputvel, sizeof(float)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(weight, &LBM_W, sizeof(float)*9));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(vectors, &a_lbmvecs, sizeof(float2)*9));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(inv, &LBM_INV, sizeof(int)*9));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(next, &a_lbmnext, sizeof(float)*9));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(ref_addr, &a_lbmref_addr, sizeof(float)*9));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(SIM_WIDTH, &a_sim_width, sizeof(int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(SIM_HEIGHT, &a_sim_height, sizeof(int)));

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_FLUID, &H_FLUID, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_NOSLIP, &H_NOSLIP, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_VELOCITY, &H_VELOCITY, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_OPENBOUNDARY_N, &H_OPENBOUNDARY_N, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_OPENBOUNDARY_S, &H_OPENBOUNDARY_S, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_OPENBOUNDARY_E, &H_OPENBOUNDARY_E, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_OPENBOUNDARY_W, &H_OPENBOUNDARY_W, sizeof(unsigned int)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(D_OPENBOUNDARY_VEL, &H_OPENBOUNDARY_VEL, sizeof(unsigned int)));
}

Simulator::~Simulator() {
	//Clean up GPU memory
	CUDA_SAFE_CALL(cudaFree(d_celltype));
	CUDA_SAFE_CALL(cudaFree(d_rho));
	CUDA_SAFE_CALL(cudaFree(d_f0));
	CUDA_SAFE_CALL(cudaFree(d_f1));
	CUDA_SAFE_CALL(cudaFree(d_u));
	CUDA_SAFE_CALL(cudaFree(d_addWind));
	

	delete [] a_ctype;
	delete [] a_f0;
	delete [] a_u;
	delete [] a_rho;

	delete [] a_lbmvecs;
	delete [] a_lbmnext;
	delete [] a_lbmref_addr;
	delete [] a_addWind;
}

/*
	Function to initialise the parameters of simulation. All data need is an image data and its width and height.
*/
unsigned int* Simulator::initialise(Image *img) {
	color pcolor;
	
	int matdim = a_sim_width * a_sim_height;

	//Wind calculation to indicates OPENBOUNDARY AND VELOCITY CELLS
	a_wind_direction = make_float2(0.f,1.f);

	//Normalitzation of wind direction
	float wind_dir_modul = sqrt( (a_wind_direction.x*a_wind_direction.x) + (a_wind_direction.y*a_wind_direction.y) );
	a_wind_direction = make_float2(a_wind_direction.x/wind_dir_modul, a_wind_direction.y/wind_dir_modul);
	
	//Calculate addition values to F becauseof wind direction
	float T = 0.f;
	for(int i=0; i<9; i++) {
		a_addWind[i] = max(dot(a_wind_direction, direction[i]), 0.f);
		T += a_addWind[i];
	}
	for(int i=0; i<9; i++) {
		a_addWind[i] = (a_addWind[i]/T) * WIND_PARTICLES;
	}

	for(int i = 0; i<a_sim_height; i++) {
		for(int j=0; j<a_sim_width; j++) {
			pcolor = img->getPixel(j,i);

			if( (pcolor.red == 255) && (pcolor.green == 255) && (pcolor.blue == 255) ) {
				a_ctype[i*a_sim_width + j] = H_FLUID;
			} else {
				a_ctype[i*a_sim_width + j] = H_NOSLIP;
			}
		}
	}

	//Modify external cells to OPENBOUNDARY
	//Bottom & Top cells
	for(int i=0; i<a_sim_width; i++) {
		a_ctype[(matdim-1)-i] = H_OPENBOUNDARY_S; //Bottom
		a_ctype[i] = H_OPENBOUNDARY_N;			//Top
	}
	//Left side & Right side cells
	for(int i=0; i<a_sim_height; i++) {
		a_ctype[i*a_sim_width] = H_OPENBOUNDARY_W;			//Left side
		a_ctype[((i+1)*a_sim_width)-1] = H_OPENBOUNDARY_E;	//Right side
	}

	//Modify neighbours of external cells if them follow the wind direction
	int init;
	if(dot(a_wind_direction,bottom_dir) > 0.f) {
		init = matdim - (2*a_sim_width) + 1;
		for(int i=0; i<(a_sim_width - 2); i++) a_ctype[init + i] = H_VELOCITY;
		for(int i=0; i<a_sim_width; i++) a_ctype[(matdim-1)-i] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,top_dir) > 0.f) {
		init = a_sim_width + 1;
		for(int i=0; i<(a_sim_width - 2); i++) a_ctype[init + i] = H_VELOCITY;
		for(int i=0; i<a_sim_width; i++) a_ctype[i] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,left_dir) > 0.f) {
		init = a_sim_width + 1;
		for(int i=0; i<(a_sim_height - 2); i++) a_ctype[init + (i*a_sim_width)] = H_VELOCITY;
		for(int i=0; i<a_sim_height; i++) a_ctype[i*a_sim_width] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,right_dir) > 0.f) {
		init = (a_sim_width - 1) * 2;
		for(int i=0; i<(a_sim_height - 2); i++) a_ctype[init + (i*a_sim_width)] = H_VELOCITY;
		for(int i=0; i<a_sim_height; i++) a_ctype[((i+1)*a_sim_width)-1] = H_OPENBOUNDARY_VEL;
	}

	CUDA_SAFE_CALL(cudaMemcpy(d_celltype, a_ctype, sizeof(unsigned int)*matdim, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_addWind, a_addWind, sizeof(float)*9, cudaMemcpyHostToDevice));
	
	//Executing GPU code
	dim3 dimBlock(BLOCKSIZE);
	dim3 dimGrid(matdim / BLOCKSIZE);
	InitLBM<<<dimGrid, dimBlock>>>(d_celltype, d_f0, d_u, d_rho);

	//Initialising values from GPU to CPU
	CUDA_SAFE_CALL(cudaMemcpy(a_f0, d_f0, sizeof(float)*matdim*9, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(a_u, d_u, sizeof(float2)*matdim, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(a_rho, d_rho, sizeof(float)*matdim, cudaMemcpyDeviceToHost));

	return a_ctype;
}

/*
	Function to realize a step over the simulation. It'll return by i/o parametres the results need to display the simulation
*/
int Simulator::nextIteration(float *velocities, float *densities) {
	int matdim = a_sim_width * a_sim_height;

	//Moving data to GPU
	CUDA_SAFE_CALL(cudaMemcpy(d_celltype, a_ctype, sizeof(unsigned int)*matdim, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_rho, a_rho, sizeof(float)*matdim, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_f0, a_f0, sizeof(float)*matdim*9, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_u, a_u, sizeof(float2)*matdim, cudaMemcpyHostToDevice));

	dim3 dimBlock(BLOCKSIZE);
	dim3 dimGrid(matdim / BLOCKSIZE);
	LBMstep<<<dimGrid,dimBlock>>>(d_celltype, d_rho, d_u, d_f0, d_f1, OMEGA, d_addWind);
	InitOpenLBM<<<dimGrid, dimBlock>>>(d_celltype, d_f1, d_u, d_rho);

	CUDA_SAFE_CALL(cudaMemcpy(a_f0, d_f1, sizeof(float)*matdim*9, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(a_u, d_u, sizeof(float2)*matdim, cudaMemcpyDeviceToHost));
	CUDA_SAFE_CALL(cudaMemcpy(a_rho, d_rho, sizeof(float)*matdim, cudaMemcpyDeviceToHost));

	//Getting return values
	for(int i=0; i<matdim; i++) {
		velocities[i*2] = a_u[i].x;
		velocities[(i*2)+1] = a_u[i].y;
		densities[i] = a_rho[i];
	}

	return 0;
}

float* Simulator::getVelocities() {
	int matdim = a_sim_width * a_sim_height;
	float *aux = new float[matdim*2];
	int iter;

	for(int i = 0; i < matdim; i++) {
		iter = i*2;
		aux[iter] = a_u[i].x;
		aux[iter+1] = a_u[i].y;
	}

	return aux;
}

float* Simulator::getDensities() {
	return a_rho;
}

float* Simulator::getForces() {
	return a_f0;
}

unsigned int* Simulator::getTypes() {
	return a_ctype;
}

void Simulator::setWindDirection(float x, float y) {
	int init_a, init_b;
	int matdim = a_sim_width * a_sim_height;
	
	a_wind_direction = make_float2(x,y);
	//Normalitzation of wind direction
	float wind_dir_modul = sqrt( (a_wind_direction.x*a_wind_direction.x) + (a_wind_direction.y*a_wind_direction.y) );
	a_wind_direction = make_float2(a_wind_direction.x/wind_dir_modul, a_wind_direction.y/wind_dir_modul);

	//Calculate addition values to F becauseof wind direction
	float T = 0.f;
	for(int i=0; i<9; i++) {
		a_addWind[i] = max(dot(a_wind_direction, direction[i]), 0.f);
		T += a_addWind[i];
	}
	for(int i=0; i<9; i++) {
		a_addWind[i] = (a_addWind[i]/T) * WIND_PARTICLES;
	}

	//Remove H_VELOCITY cells from types
	init_a = a_sim_width + 1;					//Init for top
	init_b = matdim - (2*a_sim_width) + 1;		//Init for bottom
	for(int i=0; i<(a_sim_width-2); i++) {
		a_ctype[init_a + i] = H_FLUID;
		a_ctype[init_b + i] = H_FLUID;
	}

	init_a = a_sim_width + 1;					//Init for left
	init_b = (a_sim_width - 1) * 2;				//Init for right
	for(int i=0; i<(a_sim_height-2); i++) {
		a_ctype[init_a + (i*a_sim_width)] = H_FLUID;
		a_ctype[init_b + (i*a_sim_width)] = H_FLUID;
	}

	//Remove H_OPENBOUDNARY_VEL cells from types
	//Bottom & Top cells
	for(int i=0; i<a_sim_width; i++) {
		a_ctype[(matdim-1)-i] = H_OPENBOUNDARY_S; //Bottom
		a_ctype[i] = H_OPENBOUNDARY_N;			//Top
	}
	//Left side & Right side cells
	for(int i=0; i<a_sim_height; i++) {
		a_ctype[i*a_sim_width] = H_OPENBOUNDARY_W;			//Left side
		a_ctype[((i+1)*a_sim_width)-1] = H_OPENBOUNDARY_E;	//Right side
	}

	//Calculate H_VELOCITY cells with new wind_direction
	/*init_a = a_sim_width + 1;					//Init for top
	init_b = matdim - (2*a_sim_width) + 1;		//Init for bottom
	for(int i=0; i<(a_sim_width-2); i++) {
		if(dot(a_wind_direction,top_dir) > 0.f) a_ctype[init_a + i] = H_VELOCITY;
		if(dot(a_wind_direction,bottom_dir) > 0.f) a_ctype[init_b + i] = H_VELOCITY;
	}

	init_a = a_sim_width + 1;					//Init for left
	init_b = (a_sim_width - 1) * 2;				//Init for right
	for(int i=0; i<(a_sim_height-2); i++) {
		if(dot(a_wind_direction,left_dir) > 0.f) a_ctype[init_a + (i*a_sim_width)] = H_VELOCITY;
		if(dot(a_wind_direction,right_dir) > 0.f) a_ctype[init_b + (i*a_sim_width)] = H_VELOCITY;
	}*/
	int init;
	if(dot(a_wind_direction,bottom_dir) > 0.f) {
		init = matdim - (2*a_sim_width) + 1;
		for(int i=0; i<(a_sim_width - 2); i++) a_ctype[init + i] = H_VELOCITY;
		for(int i=0; i<a_sim_width; i++) a_ctype[(matdim-1)-i] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,top_dir) > 0.f) {
		init = a_sim_width + 1;
		for(int i=0; i<(a_sim_width - 2); i++) a_ctype[init + i] = H_VELOCITY;
		for(int i=0; i<a_sim_width; i++) a_ctype[i] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,left_dir) > 0.f) {
		init = a_sim_width + 1;
		for(int i=0; i<(a_sim_height - 2); i++) a_ctype[init + (i*a_sim_width)] = H_VELOCITY;
		for(int i=0; i<a_sim_height; i++) a_ctype[i*a_sim_width] = H_OPENBOUNDARY_VEL;
	}
	if(dot(a_wind_direction,right_dir) > 0.f) {
		init = (a_sim_width - 1) * 2;
		for(int i=0; i<(a_sim_height - 2); i++) a_ctype[init + (i*a_sim_width)] = H_VELOCITY;
		for(int i=0; i<a_sim_height; i++) a_ctype[((i+1)*a_sim_width)-1] = H_OPENBOUNDARY_VEL;
	}

	//Copy data to GPU
	CUDA_SAFE_CALL(cudaMemcpy(d_celltype, a_ctype, sizeof(unsigned int)*matdim, cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpy(d_addWind, a_addWind, sizeof(float)*9, cudaMemcpyHostToDevice));
}